#include "stdafx.h"
#include <afxtempl.h>
#include "rtfconv.h"

#include "RtfElement.h"
#include "RtfDocVar.h"
#include "RtfFontInfo.h"
#include "RtfFontTable.h"
#include "RtfBookMark.h"
#include "RtfFormField.h"
#include "RtfText.h"
#include "RtfDoc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define MESSAGE(s) AfxMessageBox(s);

#ifdef _DEBUG_LOGFILE
CFile LogFile;
#endif

// RtfDoc::CRtfDocMap *RtfDoc::m_RtfDocs = 0;
CMap<CString, LPCSTR, RtfDoc*, RtfDoc*&> RtfDoc::m_RtfDocs;
//CString RtfDoc::ClarReadRecordDllName("E:\\Projects\\AdSystems\\dlls\\READREC.DLL");
CString RtfDoc::ClarReadRecordDllName("readrec.dll");
HINSTANCE RtfDoc::hInstClarionDll = 0;
typedef long (*ClarReadRecordFunc) 
	(	char *Dumm1, char *ReadResult ,
		char *Dumm2, const char*PrimFileLabel,  
		char *Dumm3, const char*PrimFilePrimKey, 
		char *Dumm4, const char*PrimFilePrimKeyVal,  
		char *Dumm5, const char*ExtraParam);
ClarReadRecordFunc CL_ReadRecord = 0;

//RtfDoc::ClarReadRecordFunc RtfDoc::CL_ReadRecord = 0;
WORD RtfDoc::CL_ReadRecordFuncNum = 2;
CString RtfDoc::m_strRtfDocID = "{\\rtf1";
CString RtfDoc::m_strRtfDocIDFakeReplacer = "      ";
CString RtfDoc::m_strPrimFileVarName("AD_PRIMARY_FILE");
char	RtfDoc::strDocVarSepChar = '@';
CString RtfDoc::DV_FilePrimKey("@PRIMKEY");
CString RtfDoc::AD_Prefix("AD_");
CString RtfDoc::DV_FilePrefix("@PREFIX");
CString RtfDoc::locFileID("FILE@");
char	RtfDoc::FieldNameSepaChar = 2;
char	RtfDoc::FieldContSepaChar = 3;
CString RtfDoc::strRtfCR(" \\par ");
char	RtfDoc::strCRSepChar = '\n';
CString RtfDoc::DV_FileChildFileSSS("@CHILDFILES");
CString RtfDoc::DV_FileParentKey("@PARENTKEY");
CString RtfDoc::DV_NrOfKeys("@NROFKEYS");
CString RtfDoc::DV_FileChildKey("@CHILDKEY");
CString RtfDoc::DV_FileUsedFields("@USEDFIELDS");
CString RtfDoc::DV_FileImpBlockSSS("@IMPBLKS");
CString RtfDoc::DV_FileImpBlock("@IMPBLK");
CString RtfDoc::AllBlockDVName("AD_ALLEXPBLKS");
CString RtfDoc::DV_FileExpBlock("AD_EXPBLK_");
CString RtfDoc::m_strPardID("\\pard");

//extern "C" long VC_OpenRtfDoc(char *a)
long VC_OpenRtfDoc(char *a)
{
	CString strFileName = a;
	return RtfDoc::OpenRtfDoc(strFileName);
}

//extern "C" void VC_CloseAllRtfDoc()
void VC_CloseAllRtfDoc()
{
	RtfDoc::CloseAllRtfDoc();
}

long VC_PrintRtfDoc(char *pcWordOutputFile, char *pcWordInputFile, char *pcPrimFileLabel, char *pcPrimFilePrimKey, char *pcPrimFilePrimKeyVal)
{	
	return RtfDoc::PrintRtfDoc(pcWordOutputFile, pcWordInputFile, pcPrimFileLabel, pcPrimFilePrimKey, pcPrimFilePrimKeyVal);
}

#ifdef _DEBUG

void RtfDoc::logFormFields()
{
	//formfields
	RtfFormField *pRtfFormField;
	CString key;
	POSITION pos = m_qRtfFormFieldQue.GetStartPosition();
	while (pos)
	{
		m_qRtfFormFieldQue.GetNextAssoc( pos, key, pRtfFormField );
		TRACE1("FormField <%s>\n", pRtfFormField->GetFieldName());
		TRACE1("FormField <%s>\n", key);
	}
}


#endif

BOOL RtfDocInitInstance()
{
	return RtfDoc::InitInstance();
}

void RtfDocExitInstance()
{
	RtfDoc::ExitInstance();
}

void RtfDoc::ExitInstance()
{

#ifdef _DEBUG_LOGFILE
	if (LogFile.m_hFile == CFile::hFileNull )
		LogFile.Close();
#endif

	if (hInstClarionDll)
	{
		FreeLibrary(hInstClarionDll);
		hInstClarionDll = 0;
		CL_ReadRecord = 0;
	}
}

BOOL RtfDoc::InitInstance()
{
#ifdef _DEBUG_LOGFILE
	LogFile.Open("c:\\rtflog.txt", CFile::modeCreate | CFile::modeWrite);
#endif

	if (hInstClarionDll && CL_ReadRecord)
		return TRUE;

	if (!hInstClarionDll)
	{
		hInstClarionDll = ::LoadLibrary(ClarReadRecordDllName);
		if (!hInstClarionDll)
		{
			DWORD err = GetLastError();
			TRACE1("error %d\n", err);
			AfxMessageBox("RtfDoc::InitInstance LoadLibrary " + ClarReadRecordDllName + " failed");
			return FALSE;
		}
	}

	DWORD funcOrd = MAKELPARAM(CL_ReadRecordFuncNum, 0);
	FARPROC fp = ::GetProcAddress(hInstClarionDll, (const char*)funcOrd);
	if (!fp)
	{
		AfxMessageBox("RtfDoc::InitInstance GetProcAddress failed");
		FreeLibrary(hInstClarionDll);
		hInstClarionDll = 0;
		return FALSE;
	}

	CL_ReadRecord  = (ClarReadRecordFunc) fp;

	return TRUE;
}

long RtfDoc::OpenRtfDoc(const CString& strFileName)
{
	CString Mess;
	Mess.Format("RtfDoc::OpenRtfDoc called,\nFile: <%s>\n", 
		(const char*) strFileName);
//	AfxMessageBox(Mess);

#ifdef _DEBUG_LOGFILE
	if (LogFile.m_hFile != CFile::hFileNull )
		LogFile.Write((const char*) strFileName, strFileName.GetLength());
#endif

	RtfDoc *aRtfDocP;
	if (m_RtfDocs.Lookup(strFileName, aRtfDocP))
	{
//		AfxMessageBox("RtfDoc::OpenRtfDoc already open");
		return 0;
	}

	aRtfDocP = new RtfDoc;
	if (aRtfDocP->Create(strFileName)) //fail
	{
		delete aRtfDocP;
		AfxMessageBox("RtfDoc::OpenRtfDoc failed");
		return 1;
	}
	else
	{
		m_RtfDocs.SetAt(strFileName, aRtfDocP);
//		AfxMessageBox("RtfDoc::OpenRtfDoc ok");
		return 0;
	}
}

void RtfDoc::CloseAllRtfDoc()
{
//	AfxMessageBox("RtfDoc::CloseAllRtfDoc called");

	if (!m_RtfDocs.GetCount())
		return;

	CString key;
	RtfDoc *value;
	POSITION pos = m_RtfDocs.GetStartPosition();
	while (pos)
	{
		m_RtfDocs.GetNextAssoc( pos, key, value );
//		AfxMessageBox("RtfDoc::CloseAllRtfDoc() a doc deleted");
		delete value;
	}
	
	m_RtfDocs.RemoveAll();
}

long RtfDoc::PrintRtfDoc(const char *pcWordOutputFile, const char *pcWordInputFile, const char *pcPrimFileLabel, const char *pcPrimFilePrimKey, const char *pcPrimFilePrimKeyVal)
{
	CString Mess;
	Mess.Format("RtfDoc::PrintRtfDoc called, \npcWordOutputFile %s\npcWordInputFile %s\npcPrimFileLabel %s\npcPrimFilePrimKey %s\npcPrimFilePrimKeyVal %s\n",
		pcWordOutputFile, 
		pcWordInputFile,
		pcPrimFileLabel,
		pcPrimFilePrimKey,
		pcPrimFilePrimKeyVal);
//	AfxMessageBox(Mess);

#ifdef _DEBUG_LOGFILE
	if (LogFile.m_hFile != CFile::hFileNull )
		LogFile.Write((const char*) Mess, Mess.GetLength());
#endif

	RtfDoc *pRtfDoc;
	if (!m_RtfDocs.Lookup(pcWordInputFile, pRtfDoc))
	{
		AfxMessageBox("RtfDoc::PrintRtfDoc not open");
		return 1;
	}

	if (pRtfDoc->Print(pcWordOutputFile, pcPrimFileLabel, pcPrimFilePrimKey, pcPrimFilePrimKeyVal))
	{
		AfxMessageBox("RtfDoc::PrintRtfDoc PrintRtfDoc failed");
		return 1;
	}
	else
		return 0;
}

RtfDoc::RtfDoc()
{
	m_RtfFontTable = 0;
	LastRtfFormField = 0;
	m_pcPrimFileReadResult = 0;
}

RtfDoc::~RtfDoc()
{
	for (int k=0; k < m_RtfQue.GetSize(); k++)
	{
		delete m_RtfQue[k];
	}

	m_RtfQue.RemoveAll();
	if (m_pcPrimFileReadResult)
		delete[] m_pcPrimFileReadResult;

//	delete m_RtfFontTable;
}

void RtfDoc::AddRtfTextToQue(CString& str)
{
	if (str.IsEmpty())
		return;

	RtfText *aRtfText = new RtfText;
	aRtfText->Create(StrBuffer);
	m_RtfQue.Add(aRtfText);


	if ( str.GetLength() == 2  && str[0] == 13 && str[1] == 10 ) //DOS-CR/LF pair
		;
	else
		PrevItemIsBookMark = FALSE;

	str.Empty();
}

void RtfDoc::ConvertDocVar()
{
	RtfDocVar *aRtfDocVar = new RtfDocVar;
	if (aRtfDocVar->Create(StrBuffer))
	{
		delete aRtfDocVar;
		throw (11);
	}

	StrBuffer.Empty();

	// docvar ok, add to general que
	m_RtfQue.Add(aRtfDocVar);
	PrevItemIsBookMark = FALSE;

	//is an uivaart docvar, so add to our search que
	if (aRtfDocVar->IsUitvaartDocVar())
	{
		const CString &DocVarName = aRtfDocVar->GetDocVarName();
		RtfDocVar *lkRtfDocVar;
		if (m_qRtfDocVarQue.Lookup(DocVarName, lkRtfDocVar))
		{
			AfxMessageBox("RtfDoc::ConvertDocVar already present in searchque");
			return;
		}

		m_qRtfDocVarQue.SetAt(DocVarName, aRtfDocVar);
	}
}

void RtfDoc::AddFormField(CString FieldName, RtfFormField *aRtfFormField)
{
	FieldName.TrimLeft();
	FieldName.TrimRight();

	RtfFormField *lkRtfFormField;
	if (m_qRtfFormFieldQue.Lookup(FieldName, lkRtfFormField))
	{
		TRACE1("RtfDoc::AddFormField %s already present in searchque\n", FieldName);
	}
	else
	{
		m_qRtfFormFieldQue.SetAt(FieldName, aRtfFormField);
	}
}

void RtfDoc::ConvertFormField()
{
	RtfFormField *aRtfFormField = new RtfFormField;
	if (aRtfFormField->Create(StrBuffer))
	{
		delete aRtfFormField;
		throw (12);
	}
	StrBuffer.Empty();

	// formfield ok, add to general que
	m_RtfQue.Add(aRtfFormField);

	CString FieldName;
	if (aRtfFormField->IsUitvaartField())
		FieldName = aRtfFormField->GetFieldName();	
	else if (!aRtfFormField->HasBookMark() && PrevItemIsBookMark )
		//then we use this as the bookmark of this field
		FieldName = LastBookMark;
	else
	{
		LastRtfFormField = aRtfFormField;
		nLastRtfFormFieldHasNoBM = TRUE;
		return;
	}

	nLastRtfFormFieldHasNoBM = FALSE;

	AddFormField(FieldName, aRtfFormField);
}

void RtfDoc::ConvertBookmark()
{
	RtfBookMark *a_RtfBookMark = new RtfBookMark;
	RtfText *aImportBMRtfText = 0;
	int nBookMarkStartPos;
	int nBookMarkEndPos;
	if (a_RtfBookMark->Create(StrBuffer, nBookMarkStartPos, nBookMarkEndPos))
	{
		delete a_RtfBookMark;
		throw (13);
	}

	if ( a_RtfBookMark->IsImportBookMark() )
	{
//TODO
		//ORG

		//it can contain multiple bookmarks
		//so que the part before the bookmark
		aImportBMRtfText = new RtfText;
		aImportBMRtfText->Create(StrBuffer.Left(nBookMarkStartPos));
		m_RtfQue.Add(aImportBMRtfText);
		// enclose this block right now
		if ( nBookMarkStartPos > 0 && StrBuffer[0] == '{' && StrBuffer.Right(1) == '}' )
		{
			aImportBMRtfText->Add('}');
			//update stringbuffer
			StrBuffer = StrBuffer.Right(StrBuffer.GetLength() - nBookMarkEndPos - 1);
			//delete the last char == '}'
			StrBuffer	= StrBuffer.Left(StrBuffer.GetLength() - 1 );
		}
		else
			//update stringbuffer
			StrBuffer = StrBuffer.Right(StrBuffer.GetLength() - nBookMarkEndPos - 1);

		//TEMP
//		StrBuffer.Empty();
	}
	else
		StrBuffer.Empty();

	// bookmark ok, add to general que
	m_RtfQue.Add(a_RtfBookMark);
	if (a_RtfBookMark->IsUitvaartBookMark() )		
	{
		if (a_RtfBookMark->IsStartBookMark() ||
			//should always try to add to the que 
			// in case of a import BM
			a_RtfBookMark->IsImportBookMark() ||
			a_RtfBookMark->IsContainingImportBookMark() ) 
		{
			const CString& bmName = a_RtfBookMark->GetBookMarkName();
			BookMarkInfo lkBookMarkInfo;
			if (m_qRtfBookMarkQue.Lookup(bmName, lkBookMarkInfo))
			{
				if ( ! a_RtfBookMark->IsImportBookMark() && !a_RtfBookMark->IsContainingImportBookMark() )
					AfxMessageBox("RtfDoc::ConvertBookmark already present in searchque");
			}
			else
			{
				lkBookMarkInfo.m_RtfBookMark = a_RtfBookMark;
				lkBookMarkInfo.m_nIndexToGeneralQue = m_RtfQue.GetUpperBound();
				m_qRtfBookMarkQue.SetAt(bmName, lkBookMarkInfo);
				if ( !a_RtfBookMark->IsImportBookMark())
				{
					PrevItemIsBookMark = TRUE;
					LastBookMark = bmName;
				}
			}
		}
		else//is endbookmark
		{
			if (nLastRtfFormFieldHasNoBM)
			{
				const CString& bmName = a_RtfBookMark->GetBookMarkName();
				AddFormField(bmName, LastRtfFormField);

				nLastRtfFormFieldHasNoBM = FALSE;
			}
		}
	}

	//TODO: ORG, no chnage needed, since StrBuffer.IsEmpty()
	if (!StrBuffer.IsEmpty())
	{
		//possible multiple bookmark
			if (StrBuffer.Find(RtfBookMark::m_strImportBlockStartBookMarkID) >= 0 || StrBuffer.Find(RtfBookMark::m_strImportBlockEndBookMarkID) >= 0 )
				ConvertBookmark();
			else
			{	//not a bookmark, so just que the rest as text to the first part before the bookmark
				//TODO: ADDED at 12/3
				aImportBMRtfText = new RtfText;
				aImportBMRtfText->Create(StrBuffer);
				m_RtfQue.Add(aImportBMRtfText);
// OLD	aImportBMRtfText->Add(StrBuffer);
				StrBuffer.Empty();
			}
	}
}

void RtfDoc::ConvertEnclosedBlock()
{
	if (!m_RtfFontTable && StrBuffer.Find(RtfFontTable::m_strFontTableMarkID) >= 0 )
	{
		m_RtfFontTable = new RtfFontTable;
      //add to the general que anyway, so it can be deleted later
		m_RtfQue.Add(m_RtfFontTable);
		if (m_RtfFontTable->Create(m_strFileName, StrBuffer))
		{
			AfxMessageBox("RtfDoc::ConvertEnclosedBlock failed");
			throw (10);
		}
		StrBuffer.Empty();
	}
	else if (StrBuffer.Find(RtfDocVar::m_strDocVarID) >= 0 )
		ConvertDocVar();
	else if (StrBuffer.Find(RtfFormField::m_strFormFieldID) >= 0 )
		ConvertFormField();
	else if (StrBuffer.Find(RtfBookMark::m_strStartBookMarkID) >= 0 ||
			StrBuffer.Find(RtfBookMark::m_strEndBookMarkID) >= 0 )
		ConvertBookmark();
	else //add everything as text
		AddRtfTextToQue(StrBuffer);
}

void RtfDoc::ParseEnclosedBlock(int &nCurFileBufferIndex)
{
	int BracketEncountered = 0;
	if (PreviousChar)
	{
        StrBuffer += PreviousChar;
        if (PreviousChar == '{') BracketEncountered++;
        PreviousChar = 0;
	}

	StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
	if (m_strFileBuffer[nCurFileBufferIndex] == '{') BracketEncountered++;
	nCurFileBufferIndex++;

	for (; nCurFileBufferIndex < m_strFileBuffer.GetLength(); nCurFileBufferIndex++)
	{
		if ( m_strFileBuffer[nCurFileBufferIndex] == '{' )
		{
			if (!PreviousChar)
			{
				BracketEncountered++;
				StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
			}
			else if (PreviousChar == '\\')
			{
				StrBuffer += PreviousChar;
				StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
				PreviousChar = 0;
			}
			else
			{
				AfxMessageBox(">StrBuffer " + StrBuffer + 
					"\nRtfDoc::ParseEnclosedBlock\nBLOCK:\nbracket open found,"+
					"but unkown previous char <<" + PreviousChar );
				PreviousChar = 0;
			}
		}
		else if ( m_strFileBuffer[nCurFileBufferIndex] == '\\' )
		{
			if (!PreviousChar)
				PreviousChar = m_strFileBuffer[nCurFileBufferIndex];
			else if ( PreviousChar == '{' || PreviousChar == '}' || PreviousChar == '\\' )
			{
				StrBuffer += PreviousChar;
				StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
				if (PreviousChar == '}' )
				{
					BracketEncountered--;
					if (!BracketEncountered)
					{
						ConvertEnclosedBlock();
						PreviousChar = 0;
						return;
					}
				}
				if ( PreviousChar == '{') BracketEncountered++;
            PreviousChar = 0;
			}
			else
			{
				AfxMessageBox(">StrBuffer " + StrBuffer + 
					"\nRtfDoc::ParseEnclosedBlock\nBLOCK:\nbracket open found,"+
					"but unkown previous char <<" + PreviousChar );
				PreviousChar = 0;
			}
		}
		else if ( m_strFileBuffer[nCurFileBufferIndex] == '}' )
		{
			if (!PreviousChar)
			{
				BracketEncountered--;
				StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
				if (!BracketEncountered)
				{
					ConvertEnclosedBlock();
					return;
				}
			}
			else if (PreviousChar == '\\' )
			{
				StrBuffer += PreviousChar;
				StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
				PreviousChar = 0;			
			}
			else
			{
				AfxMessageBox(">StrBuffer " + StrBuffer + 
					"\nRtfDoc::ParseEnclosedBlock\nBLOCK:\nbracket open found,"+
					"but unkown previous char <<" + PreviousChar );
				PreviousChar = 0;
			}
		}
		else
		{
			if (PreviousChar)
			{
				StrBuffer += PreviousChar;
				PreviousChar = 0;
			}
			StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
		}
	}

	StrBuffer.Empty();
}

void RtfDoc::ParseTextBlock(int &nCurFileBufferIndex)
{
	if (PreviousChar)
	{
		StrBuffer += PreviousChar;
		PreviousChar = 0;
	}

	StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
	nCurFileBufferIndex++;
	for (; nCurFileBufferIndex < m_strFileBuffer.GetLength(); nCurFileBufferIndex++)
	{
		if (m_strFileBuffer[nCurFileBufferIndex] == '\\')
		{
			if (PreviousChar == '\\')
			{
				StrBuffer += PreviousChar;
				StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
				PreviousChar = 0;
			}
			else if (!PreviousChar)
				PreviousChar = m_strFileBuffer[nCurFileBufferIndex];
			else
			{
				// StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
				AddRtfTextToQue(StrBuffer); //TODO ? StrBufferCounter - 1)
				PreviousChar = m_strFileBuffer[nCurFileBufferIndex];
				return;
			}
		}
		else if (m_strFileBuffer[nCurFileBufferIndex] == '{' ||
			m_strFileBuffer[nCurFileBufferIndex] == '}' )
		{
			if (PreviousChar == '\\')
			{
				StrBuffer += PreviousChar;
				StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
				PreviousChar = 0;
			}
			else
			{
				AddRtfTextToQue(StrBuffer); //TODO ? StrBufferCounter - 1)
				PreviousChar = m_strFileBuffer[nCurFileBufferIndex];
				return;
			}
		}
		else
		{
			if (PreviousChar == '\\')
			{
				StrBuffer = StrBuffer.Left(StrBuffer.GetLength() );
				AddRtfTextToQue(StrBuffer); //TODO ? StrBufferCounter - 1)

				nCurFileBufferIndex--;
				return;
			}
			else
				StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
		}
	}

	if (!StrBuffer.IsEmpty())
		AddRtfTextToQue(StrBuffer); //TODO ? StrBufferCounter - 1)
}

void RtfDoc::ParseCommandBlock(int &nCurFileBufferIndex)
{
    if (PreviousChar)
	 {
        StrBuffer += PreviousChar;
        PreviousChar = 0;
	 }

    StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
    nCurFileBufferIndex++;

	 for (; nCurFileBufferIndex < m_strFileBuffer.GetLength(); nCurFileBufferIndex++)
	 {
		if ( m_strFileBuffer[nCurFileBufferIndex] ==  '\\' ||
           m_strFileBuffer[nCurFileBufferIndex] == '{' ||
           m_strFileBuffer[nCurFileBufferIndex] == ' ' ||
           m_strFileBuffer[nCurFileBufferIndex] == '\015'
			  )
		//new block
		{
			AddRtfTextToQue(StrBuffer);
			PreviousChar = m_strFileBuffer[nCurFileBufferIndex];
			//TODO ? : a_RtfText.CreateText (StrBuffer, StrBufferCounter - 1)
			return;
		}
		else
			StrBuffer += m_strFileBuffer[nCurFileBufferIndex];
	 }

	if (StrBuffer.GetLength())
		AddRtfTextToQue(StrBuffer);
}

int RtfDoc::Write(const CString& strFileName)
{
	static CString ThisFuncName("RtfDoc::Write\n");

	CFile AnRtfFile;
//	if (!AnRtfFile.Open(strFileName, CFile::modeCreate | CFile::modeWrite | CFile::shareDenyWrite))
	if (!AnRtfFile.Open(strFileName, CFile::modeCreate | CFile::modeWrite | CFile::shareDenyNone ))
	{
		AfxMessageBox(ThisFuncName + "Open file failed:\n" + strFileName);
		return -1;
	}

	StrBuffer = m_strRtfDocID;
	int StrBufferLen;
	try
	{
	AnRtfFile.Write((const char*)StrBuffer, StrBuffer.GetLength());

    //the first entry is actually the rtf-id,
    //but we had it overwitten with blanks
	for (int k= 1; k < m_RtfQue.GetSize(); k++)
	{
		RtfElement *pRtfElement = m_RtfQue[k];
		int nRtfElemLine  = pRtfElement->GetNrOfOutputLines();
		for (int nRtfElemLineCounter=1; nRtfElemLineCounter <= nRtfElemLine; nRtfElemLineCounter++)
		{
         pRtfElement->GenerateRtfOutput(nRtfElemLineCounter, StrBuffer, StrBufferLen);
			AnRtfFile.Write((const char*)StrBuffer, StrBuffer.GetLength());
		}
	}
	}
	catch (CFileException fe)
	{
		AfxMessageBox(ThisFuncName + "write failed:\n");
		AnRtfFile.Close();
		return -2;
	}

	AnRtfFile.Close();
	return 0;
}

int RtfDoc::Print(const char *pcWordOutputFile, const char *pcPrimFileLabel, const char *pcPrimFilePrimKey, const char *pcPrimFilePrimKeyVal)
{
	static CString ThisFuncName("RtfDoc::Print\n");

	//clear this document
	ResetContent();

	StrBuffer.Empty();
	//    StrBuffer &= NEW STRING(eqRtfFileRecordSize * 10)
	//    PrimFileReadResult &= NEW CSTRING(eqChildFileReadResultBufMaxLen)

	CString PrimFilesStr;
	GetDocumentVar(m_strPrimFileVarName, PrimFilesStr);
	int NrOfPrimFiles = CountSepChar(PrimFilesStr, strDocVarSepChar);
   if (NrOfPrimFiles != 1 )
	{
		AfxMessageBox ( ThisFuncName + " err: NrOfPrimFiles ");
		return 1;
	}

    PrimFilesStr = PrimFilesStr.Left(PrimFilesStr.GetLength() -1 );
	if (PrimFilesStr != pcPrimFileLabel)
	{
		AfxMessageBox ( ThisFuncName + " Word-bestand " +
			m_strFileName + "heeft niet " + pcPrimFileLabel +
			"als hoofdbestand");
		return 1;
	}

	CString PrimKeyVarName = RtfBookMark::m_strUitvaartBookMarkID + 
		pcPrimFileLabel + DV_FilePrimKey;

	CString PrimKeyStr;
	GetDocumentVar(PrimKeyVarName, PrimKeyStr);
	if (PrimKeyStr != pcPrimFilePrimKey)
	{
		AfxMessageBox ( ThisFuncName + " Word-bestand " +
			m_strFileName + "heeft niet " + pcPrimFilePrimKey +
			"als sleutel");
	}

	CString PrimFilePrimKeyValVarName = RtfBookMark::m_strUitvaartBookMarkID +
		pcPrimFileLabel + strDocVarSepChar + pcPrimFilePrimKey;
	SetDocumentVar(PrimFilePrimKeyValVarName, pcPrimFilePrimKeyVal);

	if (!m_pcPrimFileReadResult)
		m_pcPrimFileReadResult = new char[eqChildFileReadResultBufMaxLen];
	if (!m_pcPrimFileReadResult)
	{
		AfxMessageBox ( ThisFuncName + " out of memory" );
		return 1;
	}
	memset( m_pcPrimFileReadResult, 0, eqChildFileReadResultBufMaxLen);

	char dummBlockName[2];
	dummBlockName[0] = 0;
	if ( Update(pcPrimFileLabel, m_pcPrimFileReadResult, dummBlockName, -1) < 0 )
	{
		AfxMessageBox ( ThisFuncName + " Update failed" );
		return 1;
	}

	return Write(pcWordOutputFile);
}

int RtfDoc::Create(const CString& strFileName)
{
	static CString ThisFuncName("RtfDoc::Create\n");

	CFile AnRtfFile;
	if (!AnRtfFile.Open(strFileName, CFile::modeRead | CFile::shareDenyNone))
	{
		AfxMessageBox(ThisFuncName + "Open file failed:\n" + strFileName);
		return 2;
	}
	
	char *pcFileBuffer = m_strFileBuffer.GetBuffer(AnRtfFile.GetLength());
	if (!pcFileBuffer)
	{
		AfxMessageBox(ThisFuncName + "allocating memory failed");
		return 3;
	}

	UINT nBytesRead = AnRtfFile.Read(pcFileBuffer, AnRtfFile.GetLength());
	m_strFileBuffer.ReleaseBuffer(nBytesRead);
	if ( nBytesRead < AnRtfFile.GetLength() )
	{
		AfxMessageBox(ThisFuncName + "read failed");
		return 4;
	}
	AnRtfFile.Close();

/*   prepare the variables
    CurQueCounter = 1
    CurQueEntryCounter = 1
    StrBufferCounter = 1
    PreviousChar[1] = CHR(0)
*/

	if (!ChangeRtfDocID(m_strFileBuffer))
		return 5;

	m_strFileName = strFileName;
	StrBuffer.Empty();
	PreviousChar = 0;
	PrevItemIsBookMark = FALSE;
	nLastRtfFormFieldHasNoBM = FALSE;

	try
	{
	for (int nCurFileBufferIndex = 0; nCurFileBufferIndex<m_strFileBuffer.GetLength(); nCurFileBufferIndex++)
	{
		if (m_strFileBuffer[nCurFileBufferIndex] == '{')
		{
			if (!PreviousChar)
				ParseEnclosedBlock(nCurFileBufferIndex);
			else if (PreviousChar == '{')
			{
				//fake this
				nCurFileBufferIndex--;
				PreviousChar = 0;
				ParseEnclosedBlock(nCurFileBufferIndex);				
			}
			else if (PreviousChar == '\\')
				ParseTextBlock(nCurFileBufferIndex);
			else if ( PreviousChar == ' ' || PreviousChar == '\015' )
			{
				StrBuffer += PreviousChar;
				AddRtfTextToQue(StrBuffer);
				PreviousChar = m_strFileBuffer[nCurFileBufferIndex];
			}
			else
			{
				AfxMessageBox(ThisFuncName + "PARSE:\n" + 
					"Self.m_qRtfFileQue.Rec[CurQueEntryCounter] = { AND PreviousChar unk\n" +		
					PreviousChar + "\nStrBuffer " + StrBuffer);
				break;
			}
		}
    else if (m_strFileBuffer[nCurFileBufferIndex] == '\\')
		{
			if (!PreviousChar)
				PreviousChar = m_strFileBuffer[nCurFileBufferIndex];
			else if (PreviousChar == '\\')
				ParseTextBlock(nCurFileBufferIndex);
			else if (PreviousChar == '{')
				ParseEnclosedBlock(nCurFileBufferIndex);
			else if ( PreviousChar == ' ' || PreviousChar == '\015' )
			{
				StrBuffer += PreviousChar;
				AddRtfTextToQue(StrBuffer);
				PreviousChar = m_strFileBuffer[nCurFileBufferIndex];
			}
			else
			{
				AfxMessageBox(ThisFuncName + "PARSE:\n" + 
					"Self.m_qRtfFileQue.Rec[CurQueEntryCounter] = \\ AND PreviousChar unk\n" + 
					PreviousChar + "\nStrBuffer " + StrBuffer);
				break;
			}
		}
		else
		{
			if (PreviousChar == '\\')
				ParseCommandBlock(nCurFileBufferIndex);
			else if (PreviousChar == '{')
				ParseEnclosedBlock(nCurFileBufferIndex);
			else 
				ParseTextBlock(nCurFileBufferIndex);
		}
	}//	for

	}
	catch (int)
	{
		StrBuffer.Empty();
		m_strFileBuffer.Empty();
		AfxMessageBox(ThisFuncName + " parse failed");
		return 1;
	}

	StrBuffer.Empty();
	m_strFileBuffer.Empty();
//	AfxMessageBox(ThisFuncName + " parse passed");
	return 0;
}

BOOL RtfDoc::ChangeRtfDocID(CString &strFileBuffer)
{
	int nRtfDocIDPos = strFileBuffer.Find(m_strRtfDocID);
	if (nRtfDocIDPos < 0 )
	{
		AfxMessageBox("RtfDoc::ChangeRtfDocID this is not a rtf file");
		return FALSE;
	}

	//fake this by change it to 'space'
	for (int k=0; k < m_strRtfDocIDFakeReplacer.GetLength(); k++)
	{
		strFileBuffer.SetAt(nRtfDocIDPos+k, m_strRtfDocIDFakeReplacer[k]);
	}

	return TRUE;
}

void RtfDoc::ResetContent()
{
	//bookmarks, these may have texts attached
	CString key;
	BookMarkInfo value;
	POSITION pos = m_qRtfBookMarkQue.GetStartPosition();
	while (pos)
	{
		m_qRtfBookMarkQue.GetNextAssoc( pos, key, value );
		value.m_RtfBookMark->ResetContent();
	}

	//formfields
	RtfFormField *pRtfFormField;
	pos = m_qRtfFormFieldQue.GetStartPosition();
	while (pos)
	{
		m_qRtfFormFieldQue.GetNextAssoc( pos, key, pRtfFormField );
		pRtfFormField->ResetContent();
	}

  ClearExportContent();
}

void RtfDoc::SetDocumentVar(const CString &pmDocVarName, const CString &pmDocVarValue)
{
	 RtfDocVar *pRtfDocVar;
	 if (m_qRtfDocVarQue.Lookup(pmDocVarName, pRtfDocVar))
	 {
		pRtfDocVar->SetDocVarValue(pmDocVarValue);
		return;
	 }

    //it's new, so we add it
	 CString DocVarStrBuffer = RtfDocVar::m_strDocVarID + 
		 '{' + pmDocVarName + "}{" + pmDocVarValue + "}}";
	pRtfDocVar = new RtfDocVar;
	if (pRtfDocVar->Create(DocVarStrBuffer)) //fail
	{
		AfxMessageBox("RtfDoc::SetDocumentVar create failed");
		delete pRtfDocVar;
		return;
	}

	//docvar ok, add to general que
	m_RtfQue.Add(pRtfDocVar);

	//is an uivaart docvar, so add to our search que
	if (pRtfDocVar->IsUitvaartDocVar())
		m_qRtfDocVarQue.SetAt(pmDocVarName, pRtfDocVar);
}

BOOL RtfDoc::GetDocumentVar(const CString& pmDocVarName, CString& pmDocVarValue)
{
    pmDocVarValue.Empty();

	 RtfDocVar *pRtfDocVar;
	 if (m_qRtfDocVarQue.Lookup(pmDocVarName, pRtfDocVar))
	 {
		pmDocVarValue = pRtfDocVar->GetDocVarValue();
		return TRUE;
	 }
	 else
		 return FALSE;	
}

void RtfDoc::ClearExportContent()
{
	m_qRtfTextQueExportBlock.RemoveAll();
}

int RtfDoc::CountSepChar(const char *pmStr ,char pmSepChar )
{
	int nNrOfSepChar = 0;
	do
	{
		pmStr = strchr(pmStr, pmSepChar);
		if (pmStr)
		{
			nNrOfSepChar++;
			pmStr++;
		}
	}while (pmStr);

	return nNrOfSepChar;
}

//SELF:
//	- : err
//	0 : ok
// 
int RtfDoc::Update(const char*PrimFileLabel, char *PrimFileReadResult, const char *BlockName, int ExtraInfo)
{
	static CString ThisFuncName("RtfDoc::Update\n");
//	Tracer t("RtfDoc::Update BEGIN", "RtfDoc::Update END");

	CString PrimFilePrefix;
	if (!GetDocumentVar (AD_Prefix + PrimFileLabel + DV_FilePrefix, PrimFilePrefix) )
	{
		AfxMessageBox( ThisFuncName  +
			"No prefix found for file <" + PrimFileLabel + '>');
		return -1;
	}
	
	CString PrimFilePrimKey, PrimFilePrimKeyVal;
	if (!strlen(PrimFileReadResult)) //update at the first level
	{
		GetDocumentVar(AD_Prefix + PrimFileLabel + DV_FilePrimKey, PrimFilePrimKey);
		GetDocumentVar(AD_Prefix + PrimFileLabel +  strDocVarSepChar + PrimFilePrimKey, PrimFilePrimKeyVal);
		if (PrimFilePrimKey.IsEmpty() || PrimFilePrimKeyVal.IsEmpty())
		{
			AfxMessageBox( ThisFuncName  + "Key or Keyval not set");
			return -2;
		}

		ReadRecord(PrimFileReadResult, PrimFileLabel, 
			PrimFilePrimKey, PrimFilePrimKeyVal, "SET");
	}

	CString WordBMPrefix;
	GetWordBMPrefix(PrimFilePrefix, WordBMPrefix);
	if (UpdateFields(PrimFileLabel, 
        AD_Prefix + PrimFilePrefix,
        AD_Prefix + WordBMPrefix + '_',
        PrimFileReadResult))
	{
			AfxMessageBox( ThisFuncName  + "UpdateFields failed");
			return -1;
	}

//UpdateChildBlock
// - : no child
//	- : err
// 0 : ok


//UpdateImportBlock
// 0 : no import block
//	0 : finished
// - : err


//UpdateExportBlock 
// - : err
//	UpdateChildBlockRes

//SELF:
//	- : err
//	0 : ok

//   IMPORT BLOCK, next release
    int UpdateImportBlockRes = UpdateImportBlock ( 
		 PrimFileLabel, PrimFileReadResult,
		 BlockName, ExtraInfo);
    //result of UpdateImportBlockRes is not important
	
//   EXPORT BLOCK
    if (strlen(BlockName))
	 {
        int UpdateExportBlockRes = UpdateExportBlock ( 
				PrimFileLabel, PrimFileReadResult,
				BlockName, ExtraInfo);
        return UpdateExportBlockRes;
    }

//   CHILD FILES
	 CString ChildFileStr;
    GetDocumentVar(AD_Prefix + PrimFileLabel + 
		 DV_FileChildFileSSS, ChildFileStr);
    if (ChildFileStr.GetLength())
	 {
        int UpdateChildBlockRes = UpdateChildBlock(PrimFileLabel,PrimFileReadResult, ChildFileStr, ExtraInfo);
        return UpdateChildBlockRes;
    }

	return 0;
}

int DEBUG_ReadRecord = -1;
void RtfDoc::ReadRecord( char *ReadResult, const char *PrimFileLabel, const char *PrimFilePrimKey,  const char *PrimFilePrimKeyVal,  const char *ExtraParam)
{
//	if (++DEBUG_ReadRecord)
//		return;

	CString Mess;
/*	Mess.Format("RtfDoc::ReadRecord\nReadResult %s\nPrimFileLabel %s\nPrimFilePrimKey %s\nPrimFilePrimKeyVal %s\nExtraParam %s\n",
		ReadResult,
		PrimFileLabel,
		PrimFilePrimKey,
		PrimFilePrimKeyVal,
		ExtraParam);
	AfxMessageBox(Mess); */

#ifdef _DEBUG_LOGFILE
	if (LogFile.m_hFile != CFile::hFileNull )
		LogFile.Write((const char*) Mess, Mess.GetLength());
#endif

	/*
	CFile inpFile;
	inpFile.Open("d:\\rtflog.txt", CFile::modeRead | CFile::shareDenyNone);
	memset(ReadResult, 0, eqChildFileReadResultBufMaxLen);
	inpFile.Read(ReadResult, inpFile.GetLength());
	inpFile.Close();
	return;
*/
//	int oldLen = strlen(ReadResult);
	static char dummParam[] ="NOP";
	if (hInstClarionDll && CL_ReadRecord)
		CL_ReadRecord(
			dummParam, ReadResult, 
			dummParam, PrimFileLabel, 
			dummParam, PrimFilePrimKey,  
			dummParam, PrimFilePrimKeyVal,  
			dummParam, ExtraParam);
	else
		AfxMessageBox("CL_ReadRecord not set");

//	Mess.Format("Before has len %d\nAfter has len %d", oldLen, strlen(ReadResult));
//	AfxMessageBox(Mess);

	Mess = "\n\n ********* NEWBLOCK ********* \n\n";

#ifdef _DEBUG_LOGFILE
	if (LogFile.m_hFile != CFile::hFileNull )
		LogFile.Write((const char*) Mess, Mess.GetLength());
	if (LogFile.m_hFile != CFile::hFileNull )
		LogFile.Write(ReadResult, strlen(ReadResult));
#endif
}

void RtfDoc::GetWordBMPrefix( const CString &aClarionPrefix, CString& WordBMPrefix)
{
	GetDocumentVar("AD_PR_" + aClarionPrefix, WordBMPrefix);
}

int RtfDoc::UpdateFields ( CString FileLabel,  CString FilePrefix,  CString WordBMPrefix,  char *RecordContent)
{
	static CString ThisFuncName("RtfDoc::UpdateFields\n");
   if (strstr(RecordContent, (const char*) locFileID) != RecordContent)
	{
		CString Mess;
		AfxMessageBox( ThisFuncName  + "locFileID not foudn, not a file info str");
		Mess.Format("RecordContent has len %d", strlen(RecordContent) );
		AfxMessageBox( Mess);
		return 1;
	}

	 CString strRecordContent(RecordContent);
	int FileLabelEndPos = strRecordContent.Find(FieldNameSepaChar);
	int locFileIDLen = locFileID.GetLength();

	CString FileLabelCheck = strRecordContent.Mid(
		locFileIDLen,
		FileLabelEndPos - locFileIDLen); //TODO See orig
	if (FileLabel != FileLabelCheck)
	{
		AfxMessageBox( ThisFuncName  + "Wrong file read" +
			"needed FileLabel " + FileLabel + 
		"\nactually read  FileLabel " + FileLabelCheck );
		return 2;
	}
    int FieldLabelSP = FileLabelEndPos + 1;

	 int FieldLabelEP;
	 int FieldValSP;
	 int FieldValEP;
	 while(FieldLabelSP)
	 {
		FieldLabelEP = strchr(RecordContent + FieldLabelSP, FieldNameSepaChar) - RecordContent;
		if (FieldLabelEP < 0 )
			return 0;
		FieldValSP   = FieldLabelEP  + 1;
      FieldValEP   = strchr(RecordContent + FieldValSP, FieldContSepaChar) - RecordContent;
		if (FieldValEP < 0 || FieldLabelEP < 0)
            return 0; //everything is updated now

		CString FieldLabel = strRecordContent.Mid( FieldLabelSP,
			FieldLabelEP - FieldLabelSP);
		CString FieldVal = strRecordContent.Mid(FieldValSP, FieldValEP - FieldValSP);
		CString BMField = WordBMPrefix + FieldLabel;
		if (BMField.GetLength() > eqWordBookMarkMaxLen)
			BMField = BMField.Left(eqWordBookMarkMaxLen);

		int FieldValLen = FieldVal.GetLength();
		SetFormResult(BMField, FieldVal , FieldValLen);
		FieldLabelSP = FieldValEP + 1;
	 }

	return 0;
}

void RtfDoc::SetFormResult (const CString &pmFieldName, const CString &pmFieldValue, int pmFieldValueLen)
{
	RtfFormField *pRtfFormField;
	if (m_qRtfFormFieldQue.Lookup(pmFieldName, pRtfFormField) )
	{
		CString TargetStr;
		ConvertCR(TargetStr, pmFieldValue);
		pRtfFormField->SetFieldValue(TargetStr);
	}
}

void RtfDoc::ConvertCR(CString &TargetStr, const CString &SrcStr)
{
	TargetStr.Empty();
	for (int k=0; k < SrcStr.GetLength(); k++)
	{
		if (SrcStr[k] == strCRSepChar )
			TargetStr += strRtfCR;
		else
			TargetStr += SrcStr[k];
	}
}

// 0 : no import block
//	0 : finished
// - : err
int RtfDoc::UpdateImportBlock ( CString PrimFileLabel, char* PrimFileReadResult, CString BlockName, int ExtraInfo)
{
	static CString ThisFuncName("RtfDoc::UpdateImportBlock\n");
//	Tracer t("RtfDoc::UpdateImportBlock BEGIN", "RtfDoc::UpdateImportBlock END");

	CString AllImpBlkDVName = AD_Prefix + PrimFileLabel + DV_FileImpBlockSSS;
	CString AllImpBlkDV, OtherImpBlkName;
	GetDocumentVar(AllImpBlkDVName, AllImpBlkDV);
	int NrOfOtherBlocks = CountSepChar(AllImpBlkDV, strDocVarSepChar);
	if (NrOfOtherBlocks <= 0)
        return 0;

	//Clear old content of the blocks
    int PrevImpBlkCrPos = 0;
    int NexImpBlktCrPos = 0;	
	CArray<CString, LPCSTR>  OtherBlksShortAr;
    for (int OtherBlkIndex = 1; OtherBlkIndex <= NrOfOtherBlocks; OtherBlkIndex++)
	 {
        NexImpBlktCrPos = strchr( (const char*) AllImpBlkDV + PrevImpBlkCrPos, strDocVarSepChar) - (const char*) AllImpBlkDV;
		  OtherBlksShortAr.Add(AllImpBlkDV.Mid(PrevImpBlkCrPos, NexImpBlktCrPos -PrevImpBlkCrPos));
        ClearBlock("AD_IMPBLK_" + 
            OtherBlksShortAr[OtherBlksShortAr.GetUpperBound()] +  "_B");
        PrevImpBlkCrPos = NexImpBlktCrPos + 1;
	 }

	//update the blocks
	for (OtherBlkIndex =  1 ; OtherBlkIndex <= NrOfOtherBlocks; OtherBlkIndex++)
	{
		const CString & strOtherBlksShort = OtherBlksShortAr[OtherBlkIndex-1];
		BookMarkInfo aBookMarkInfo;
		if (!m_qRtfBookMarkQue.Lookup("AD_IMPBLK_" + strOtherBlksShort + "_B", aBookMarkInfo))
		{
        MESSAGE ( ThisFuncName + 
          "can not find bookmark for import")
          return -1;
      }
      RtfBookMark *ImpBlkBookMark = aBookMarkInfo.m_RtfBookMark;
		CString OtherImpBlkFileName;
      GetDocumentVar( AD_Prefix + PrimFileLabel + DV_FileImpBlock + 
            '_' + strOtherBlksShort + "_FILENAME", 
            OtherImpBlkFileName );
        if (OtherImpBlkFileName.IsEmpty())
		  {
            MESSAGE ( ThisFuncName + 
                "Getting docvar for IMPBLK_FILENAME failed")
            return -1;
			}

        OtherImpBlkFileName = RtfToDosPath(OtherImpBlkFileName);
        if (BlkImpFileOpen(OtherImpBlkFileName))
		  {
            MESSAGE ( ThisFuncName + 
                "opening " + OtherImpBlkFileName + " failed")
            return -1;
        } 

		  RtfDoc *pRtfDoc;
		  if (!m_RtfDocs.Lookup(OtherImpBlkFileName, pRtfDoc)) 
			//not opened yet, shouldn't happen here
		  {
            MESSAGE ( ThisFuncName +
                OtherImpBlkFileName +" not open yet, fatal err")
            return -1;
		  }
//        !merge the fonttable
        m_RtfFontTable->MergeFontTable( *(pRtfDoc->m_RtfFontTable));
        GetDocumentVar(AD_Prefix + PrimFileLabel + DV_FileImpBlock +
            '_' + strOtherBlksShort + "_NAME",
            OtherImpBlkName);

		  int UpdRes = 0;
        int ImpBlkUpdCounter = 0;
        while( !UpdRes)
		  {
            UpdRes = pRtfDoc->Update(PrimFileLabel, 
                    PrimFileReadResult,
                    OtherImpBlkName, ImpBlkUpdCounter);
            if (!UpdRes)
              CopyImportBlocks(ImpBlkBookMark, pRtfDoc);

            ImpBlkUpdCounter++;
        }//Wend UpdRes > 0

	}
	return 0;
}

void RtfDoc::TranslateFontNr(CString &strImpblockCopy, RtfDoc *pRtfDoc)
{
	int locFontIdStartPos = 0;
	int locOldBufLastPos = 0;
	CString m_strImportBlockBuffNew;
	int locOldBufLen = strImpblockCopy.GetLength();
	while(TRUE)
	{
		locFontIdStartPos = strstr( (const char*) strImpblockCopy + locOldBufLastPos, "\\f") - (const char*) strImpblockCopy;
	    if (locFontIdStartPos < 0) //no more font info
		 {
			int locCharsToCopy = locOldBufLen - locOldBufLastPos + 1;
			m_strImportBlockBuffNew += strImpblockCopy.Right(locOldBufLen - locOldBufLastPos);
			strImpblockCopy = m_strImportBlockBuffNew;
			return;
		 }

		locFontIdStartPos += 1; //LEN('\f') - 1
		 m_strImportBlockBuffNew += strImpblockCopy.Mid(locOldBufLastPos, locFontIdStartPos - locOldBufLastPos + 1 );
		locOldBufLastPos += (locFontIdStartPos - locOldBufLastPos + 1);

	    if (!isdigit(strImpblockCopy[locFontIdStartPos+ 1]))
		 //!not a font info
			continue;

		CString locStrOldFontNr;
		for ( int locFontIdCnt = locFontIdStartPos + 1;
			locFontIdCnt < locOldBufLen;
			locFontIdCnt++)
		{
			if (isdigit(strImpblockCopy[locFontIdCnt] ))
			  locStrOldFontNr += strImpblockCopy[locFontIdCnt];
			else
			  break;
		}

		int locOldFontNr = atoi( (const char*) locStrOldFontNr);
		int locNewFontNr = m_RtfFontTable->GetMappedFontNr (pRtfDoc->m_strFileName, locOldFontNr);
		if (locNewFontNr < 0 )
		{
			MESSAGE ( "RtfDoc::TranslateFontNr " 
			"getting locNewFontNr failed, set it to old one")
			locNewFontNr = locOldFontNr;
		}

    CString locStrNewFontNr;
	 locStrNewFontNr.Format("%1d", locNewFontNr);

    locOldBufLastPos = locFontIdCnt;
    m_strImportBlockBuffNew += locStrNewFontNr;
	}
}

void RtfDoc::CopyImportBlocks(RtfBookMark *ImpBlkBookMark, RtfDoc *pRtfDoc)
{
	for (int srcExportBlkCnt = 0;
		srcExportBlkCnt < pRtfDoc->m_qRtfTextQueExportBlock.GetSize();
		srcExportBlkCnt++)
	{
		CString strImpblockCopy = pRtfDoc->m_qRtfTextQueExportBlock[srcExportBlkCnt];
		TranslateFontNr(strImpblockCopy, pRtfDoc);
		ImpBlkBookMark->AddRtfText(strImpblockCopy);
	}

	pRtfDoc->m_qRtfTextQueExportBlock.RemoveAll();
}

// - : err
//	UpdateChildBlockRes
int RtfDoc::UpdateExportBlock ( CString PrimFileLabel, char* PrimFileReadResult, CString BlockName, int ExtraInfo)
{
	static CString ThisFuncName("RtfDoc::UpdateExportBlock\n");
//	Tracer t("RtfDoc::UpdateExportBlock BEGIN", "RtfDoc::UpdateExportBlock END");

	CString AllBlockDV;
	GetDocumentVar(AllBlockDVName, AllBlockDV);
	int NrOfExpBlock = CountSepChar(AllBlockDV, strDocVarSepChar);
   if (NrOfExpBlock <= 0) return -1;

	CString ExpBlockNameStr;
	CString ExpBlockShortStr;
	int PrevExpBlkCrPos = 0;
   int NextExpBlkCrPos = 0;
	for (int ExpBlockIndex = 1;
		ExpBlockIndex <= NrOfExpBlock;
		ExpBlockIndex++)
	{
        NextExpBlkCrPos = strchr( (const char*) AllBlockDV + PrevExpBlkCrPos, strDocVarSepChar) - (const char*) AllBlockDV;
        ExpBlockShortStr = AllBlockDV.Mid(PrevExpBlkCrPos, NextExpBlkCrPos - PrevExpBlkCrPos);

        GetDocumentVar(DV_FileExpBlock + ExpBlockShortStr+ "_NAME" ,
            ExpBlockNameStr);
        if (ExpBlockNameStr == BlockName)
            break;

        PrevExpBlkCrPos = NextExpBlkCrPos + 1;
	}//! LOOP ExpBlockIndex

   if (ExpBlockNameStr != BlockName)
	{
        MESSAGE ( ThisFuncName +
            " Geen geschikte blok gevonden" +
            "BlockName " + BlockName )
        return -1;
	}

	CString ExpBlkPrimFile;
	GetDocumentVar(DV_FileExpBlock + ExpBlockShortStr + "_PRIMF",
        ExpBlkPrimFile);
    if ( ExpBlkPrimFile != PrimFileLabel)
	 {
        MESSAGE ( ThisFuncName +
            "De hoofdbestanden zijn niet hetzelfde" +
            "ExpBlkPrimFile " + ExpBlkPrimFile + 
            "PrimFileLabel " + PrimFileLabel)
        return -1;
	 }

	CString ChildFileStr;
	GetDocumentVar(DV_FileExpBlock + ExpBlockShortStr + "_CHILDF",
        ChildFileStr);
    int UpdateChildBlockRes = UpdateChildBlock(PrimFileLabel,PrimFileReadResult, ChildFileStr, ExtraInfo);
    if (UpdateChildBlockRes < 0 )
	 {
//      CLEAR(Self.m_strExportBlockBuff) // !just for sure
      return -1;
	 }

//!here should paste the updated block to a buffer, so it can
//!be copied by the caller

	CString BMExpBlockBegin = "AD_EXPBLOCK_" + ExpBlockShortStr + "_BP";
	CString BMExpBlockEnd   = "AD_EXPBLOCK_" + ExpBlockShortStr + "_EP";
	BookMarkInfo aBookMarkInfo;
	if (!m_qRtfBookMarkQue.Lookup(BMExpBlockBegin, aBookMarkInfo))
	{
		MESSAGE ( ThisFuncName +
			"can not find bookmark " + BMExpBlockBegin + " for export" + 
         "in file " + m_strFileName)
		return -1;
	}
   int BMExpBlockBeginIndex = aBookMarkInfo.m_nIndexToGeneralQue;

	if (!m_qRtfBookMarkQue.Lookup(BMExpBlockEnd, aBookMarkInfo))
	{
		MESSAGE ( ThisFuncName +
			"can not find bookmark " + BMExpBlockEnd + " for export" + 
         "in file " + m_strFileName)
		return -1;
	}
	int BMExpBlockEndIndex = aBookMarkInfo.m_nIndexToGeneralQue;

	if (	BMExpBlockBeginIndex > m_RtfQue.GetSize() ||
			BMExpBlockEndIndex > m_RtfQue.GetSize() ||
			BMExpBlockEndIndex - BMExpBlockBeginIndex <= 1)
	{
		MESSAGE ( ThisFuncName + 
            "index for bookmark out of range")
          return -1;
	}

//  !search for the command with the last tab-layout
  int BMExpBlockBeginIndexBackWard = BMExpBlockBeginIndex - 1;
  int nWidthControlBegin = 0;
  while(BMExpBlockBeginIndexBackWard > 0 )
  {
    if (m_RtfQue[BMExpBlockBeginIndexBackWard]->IsWidthControlID())
	 {
      nWidthControlBegin = BMExpBlockBeginIndexBackWard;
      break;
	 }

    BMExpBlockBeginIndexBackWard--;
  }

  int nWidthControlEnd = 0;
  if (nWidthControlBegin)
  {
    for (int nWidthControlIndex = nWidthControlBegin + 1;
					nWidthControlIndex < BMExpBlockBeginIndex;
					nWidthControlIndex++)      
		if (!m_RtfQue[nWidthControlIndex]->IsWidthControlArgumentID())
		{
        nWidthControlEnd = nWidthControlIndex - 1;
        break;
		}
  }

//  !this will also take the tab-layout with us
	if (nWidthControlBegin && nWidthControlEnd)
	{
		for (int i = nWidthControlBegin;	i <= nWidthControlEnd; i++)
		{
			CString expStr;
			int expStrLen;
			m_RtfQue[i]->GenerateRtfOutput(1, expStr, expStrLen);

			if (i == nWidthControlEnd) // add a space after the last control-text Self.m_strExportBlockBuff = CLIP(Self.m_strExportBlockBuff) & ' '
	        expStr += ' ';

			m_qRtfTextQueExportBlock.Add(expStr);
		}
	}

//!now find the tab-controls at the end of the exportblock, it shouldn't be
//!copied
	int TailTabIndexBegin = 0;
	int TailTabIndexEnd   = 0;
	int TailTabIndexCounter = BMExpBlockEndIndex;
	while(TRUE)
	{
		if (TailTabIndexCounter < BMExpBlockBeginIndex) break;
		RtfElement *pRtfElement = m_RtfQue[TailTabIndexCounter];

		if (!TailTabIndexEnd) //!end not found
		{
			if (pRtfElement->IsFormField()) break;
			if (pRtfElement->IsWidthControlArgumentID())
			{
				TailTabIndexEnd = TailTabIndexCounter;
				continue;
			}
		}
		else if (!TailTabIndexBegin) // begin not found
		{
			if (pRtfElement->IsFormField()) break;
			if (pRtfElement->IsWidthControlID())
			{
				TailTabIndexBegin = TailTabIndexCounter;
				break;
			}
		}
		else
		{
			if (pRtfElement->IsFormField())
			{
				TailTabIndexBegin = 0;
				TailTabIndexEnd   = 0;
				break;
			}
		}

		TailTabIndexCounter--;
	}

	if (!TailTabIndexBegin || !TailTabIndexEnd)
	{
		TailTabIndexBegin = 0;
		TailTabIndexEnd   = 0;
	}

	int LastCharIsSpace = 0;
	int LastCharIsAccoladeSluiten = 0;
	CString expStrOld;
	bool bLastStrIsPardID = false;

	//solution for whatether copy the stuff together with the
	//begin-bookmark of a exportblock
	if ( !(m_RtfQue[BMExpBlockBeginIndex+1]->IsFormField()))
	{
		CString expStr;
		int expStrLen;
		m_RtfQue[BMExpBlockBeginIndex]->GenerateRtfOutput(1, expStr, expStrLen);
		m_qRtfTextQueExportBlock.Add(expStr);
	}

	for (int I= BMExpBlockBeginIndex + 1; I < BMExpBlockEndIndex; I++)
//	for (int I= BMExpBlockBeginIndex ; I < BMExpBlockEndIndex; I++)
	{
		if (TailTabIndexBegin <= I && I <= TailTabIndexEnd) 
			continue;

		RtfElement *pRtfElement = m_RtfQue[I];

		CString expStr;
		int expStrLen;
		pRtfElement->GenerateRtfOutput(1, expStr, expStrLen);

		//do not copy the \pard
    if (expStr.Find(m_strPardID) >= 0)
		{
			bLastStrIsPardID = true;
			continue;
		}

		// since we leave some stuff like \pard, 
		// the space before and after the \pard, 
		// example: ' \pard ' , will cause two spaces, 
		// this should be prevented
		if (IsSpace(expStr))
		{
			if ( bLastStrIsPardID )
			{
				LastCharIsSpace += 1;
				if (LastCharIsAccoladeSluiten && LastCharIsSpace == 1)
				{
					LastCharIsSpace = 0;
					continue;
				} //!space char right after right-accolade

				// there were already space-char copied
				if (LastCharIsSpace > 1) 
					continue;

				//the last character of the previous string was a space-char
				if ( IsSpace(expStrOld.Right(1)) )
					continue;

				expStr = expStr.Left(1);
			}
		}
    else // NOT (IsSpace(expStr))
		{
			bLastStrIsPardID = false;

			if (!expStr.IsEmpty() && ( expStr[expStr.GetLength() - 1 ] == ' ') )
				LastCharIsSpace = 1;
			else if (!expStr.IsEmpty() && ( expStr[0] == ' ') 
								&& !expStrOld.IsEmpty() 
								&& expStrOld[expStrOld.GetLength() - 1 ] == ' ')
				expStr = expStr.Right(expStr.GetLength() - 1);
			else
				LastCharIsSpace = 0; //!reset the counter
		}	//fi IsSpace(expStr))

		m_qRtfTextQueExportBlock.Add(expStr);

    if (expStr.GetLength() && expStr[expStr.GetLength() - 1 ] == '}')
			LastCharIsAccoladeSluiten = 1;
    else
			LastCharIsAccoladeSluiten = 0;

		expStrOld = expStr;
	}//end for : int I

	//new copy the stuff from the end-bookmark, for the case it contains
	//additional info
	RtfBookMark *pRtfBookMark = (RtfBookMark*) m_RtfQue[I];
	if ( pRtfBookMark->IsUitvaartExportBM() )
	{
		CString str;
		pRtfBookMark->GenerateTextBeforeBM(str);
		if (!str.IsEmpty())
			m_qRtfTextQueExportBlock.Add(str);
	}
	return UpdateChildBlockRes;
}

//UpdateChildBlock
// - : no child
//	- : err
// 0 : ok
int RtfDoc::UpdateChildBlock ( CString PrimFileLabel, char *PrimFileReadResult, CString BlockName, int ExtraInfo)
{
	static CString ThisFuncName("RtfDoc::UpdateChildBlock\n");
//	Tracer t("RtfDoc::UpdateChildBlock BEGIN", "RtfDoc::UpdateChildBlock END");

    int NrOfChildFiles = CountSepChar(BlockName, strDocVarSepChar);
    if (NrOfChildFiles  <= 0 )
        return -1;

	CArray<CString, LPCSTR> ChildFileAr;
    int PrevChildCrPos = 0;
    int NextChildCrPos = 0;
    for (int ChildFileIndex = 1; ChildFileIndex <= NrOfChildFiles; ChildFileIndex++)
	 {
        NextChildCrPos = strchr( (const char*) BlockName + PrevChildCrPos, strDocVarSepChar) - (const char*) BlockName;
		  if (NextChildCrPos < 0 )
			  return -1;

        ChildFileAr.Add(BlockName.Mid(PrevChildCrPos, NextChildCrPos - PrevChildCrPos));
        PrevChildCrPos = NextChildCrPos + 1;
    }

	char *ChildFileReadResult = new char[eqChildFileReadResultBufMaxLen];
	memset( ChildFileReadResult, 0, eqChildFileReadResultBufMaxLen);
	CString ChildFilePrefix,ParentKey, PrimFilePrefix, ParentKeyField;
	CString ChildKey, ExtraReadRecordParam;
	CString ChildWordBMPrefix, ParentWordBMPrefix;
	int ParentKeyFieldSP, ParentKeyFieldEP;
	try
	{
	for (ChildFileIndex = 1; ChildFileIndex <= NrOfChildFiles; ChildFileIndex++)
	{
		const CString &ChildFileRef = ChildFileAr[ChildFileIndex - 1];
		GetDocumentVar(AD_Prefix + ChildFileRef +
			DV_FilePrefix, ChildFilePrefix);
		if (ChildFilePrefix.IsEmpty())
		{
			AfxMessageBox(ThisFuncName +
                "missing child file prefix for file " +
                ChildFileRef);
			throw -1;
		}

		GetDocumentVar(AD_Prefix + PrimFileLabel +
            strDocVarSepChar + ChildFileRef + 
            DV_FileParentKey, ParentKey);
		if (ParentKey.IsEmpty())
		{
			AfxMessageBox(ThisFuncName +
                "missing parent-key for file " +
                ChildFileRef + 
					"with parent file " + PrimFileLabel);
			throw -1;
		}

		GetDocumentVar (AD_Prefix + PrimFileLabel + DV_FilePrefix, PrimFilePrefix);
      ParentKeyField = ParentKeyToField(PrimFilePrefix, ParentKey);
		ParentKeyFieldSP = strstr(PrimFileReadResult, (const char*)(ParentKeyField + FieldNameSepaChar)) - PrimFileReadResult;
		if (ParentKeyFieldSP < 0 )
		{
			AfxMessageBox(ThisFuncName +
                "Can not find find matching parent key(ParentKeyFieldSP) in the read data"+
                "to be find:"
                "\nParentKey " + ParentKey + 
                "\nwith field " + ParentKeyField);
			throw -1;
		}
		ParentKeyFieldSP += ParentKeyField.GetLength() + 1;
      ParentKeyFieldEP = strchr(PrimFileReadResult + ParentKeyFieldSP, FieldContSepaChar) - PrimFileReadResult;
		if (ParentKeyFieldEP < 0 || (ParentKeyFieldEP - ParentKeyFieldSP <= 1) )
		{
			AfxMessageBox(ThisFuncName +
                "Can not  find matching parent key(ParentKeyFieldEP)  in the read data" +
                "to be find:" +
                "\nParentKey " + ParentKey + 
                "\nwith field " + ParentKeyField );
			throw -1;            
		}

		CString strPrimFileReadResult(PrimFileReadResult);
		CString ParentKeyVal = strPrimFileReadResult.Mid(ParentKeyFieldSP, ParentKeyFieldEP - ParentKeyFieldSP);
		GetDocumentVar(AD_Prefix + PrimFileLabel +
            strDocVarSepChar + ChildFileRef +
            DV_FileChildKey, ChildKey);
		if (ChildKey.IsEmpty())
		{
			AfxMessageBox(ThisFuncName + "missing child-key");
			throw -1;
		}

		if (ExtraInfo >= 0)
			ExtraReadRecordParam.Format("LOOP@%1d", ExtraInfo);
		else
			ExtraReadRecordParam = "SET";

		ReadRecord (ChildFileReadResult, ChildFileRef,
			ChildKey, ParentKeyVal, ExtraReadRecordParam);

		GetWordBMPrefix(ChildFilePrefix, ChildWordBMPrefix);
      GetWordBMPrefix(PrimFilePrefix, ParentWordBMPrefix);

		if (!strlen(ChildFileReadResult))
		{
			if (ExtraReadRecordParam =="SET") //it's ok to have empty result
			{
				ClearContent(AD_Prefix + PrimFileLabel + 
					strDocVarSepChar + ChildFileRef,
               AD_Prefix + PrimFilePrefix + '_' + ChildFilePrefix + '_',
               AD_Prefix + ParentWordBMPrefix + '_' +
               ChildWordBMPrefix + '_');
			}
			else//ExtraReadRecordParam was= 'LOOP', so we return a -1
				throw -1; //to indicate that there are no more blocks to be updated
		}
		else
		{
			UpdateFields(ChildFileRef,
				AD_Prefix + PrimFilePrefix + '_' + ChildFilePrefix+ '_',
            AD_Prefix + ParentWordBMPrefix+ '_' +
            ChildWordBMPrefix + '_', 
            ChildFileReadResult);
		}
	}
	}
	catch(int errNr)
	{
		delete[] ChildFileReadResult;
		return errNr;
	}

	delete[] ChildFileReadResult;
	return 0;
}

CString RtfDoc::ParentKeyToField ( const CString &PrimFilePrefix, const CString &ParentKey)
{
	CString Res;

	CString strNrOfKeys;
	GetDocumentVar(AD_Prefix + PrimFilePrefix + DV_NrOfKeys,
        strNrOfKeys);
	int NrOfKeys = atoi( (const char*) strNrOfKeys);
	if (NrOfKeys <= 0)
	{
		AfxMessageBox("RtfDoc::ParentKeyToField  NrOfKeys not found or illegal");
		return Res;
	}

	CString KeyName, KeyField;
	for (int KeyIndex = 1; KeyIndex <= NrOfKeys; KeyIndex++)
	{
		CString dumm;
		dumm.Format("%s%s@KEY%1d",
			(const char*) AD_Prefix,
			(const char*) PrimFilePrefix,
         KeyIndex);
		GetDocumentVar( dumm, KeyName);
		if (KeyName == ParentKey)
		{
			dumm.Format("%s%s@KEY%1dFIELDS",
				(const char*) AD_Prefix,
				(const char*) PrimFilePrefix,
				KeyIndex);
			GetDocumentVar( dumm, KeyField);
			if (KeyField.IsEmpty())
			{
				AfxMessageBox("RtfDoc::ParentKeyToField KeyField not found");
				return Res;
			}

			return KeyField;
		}
	}

	return Res;
}

void RtfDoc::ClearContent(CString  FileLabel, CString FilePrefix, CString WordBMPrefix)
{
	static CString ThisFuncName ("RtfDoc::ClearContent\n");

	CString FieldStr;
	GetDocumentVar(FileLabel + DV_FileUsedFields, FieldStr);
	int NrOfChildFileFields = CountSepChar(FieldStr, strDocVarSepChar);
	if (NrOfChildFileFields <= 0)
	{
        AfxMessageBox(ThisFuncName + 
            "no ChildFileFields, empty");
        return;
	}

    int PrevChildFileFieldSTXPos = 0;
    int NextChildFileFieldSTXPos = 0;
	CString FieldValue; //empty dumm val

    for (int ChildFileFieldIndex = 1; ChildFileFieldIndex <= NrOfChildFileFields; ChildFileFieldIndex++)
	 {
        NextChildFileFieldSTXPos = strchr((const char*)FieldStr + PrevChildFileFieldSTXPos, strDocVarSepChar) - (const char*)FieldStr;

        CString FieldLabel = FieldStr.Mid(PrevChildFileFieldSTXPos, NextChildFileFieldSTXPos - PrevChildFileFieldSTXPos);
        PrevChildFileFieldSTXPos = NextChildFileFieldSTXPos + 1;

        CString BMField = WordBMPrefix + FieldLabel;
        if (BMField.GetLength() > eqWordBookMarkMaxLen)
            BMField = BMField.Left(eqWordBookMarkMaxLen);

        SetFormResult(BMField, FieldValue, 0);
	 }
}

void RtfDoc::ClearBlock (CString  aBMImpBlkBegin)
{
	BookMarkInfo aBookMarkInfo;
	if (m_qRtfBookMarkQue.Lookup(aBMImpBlkBegin, aBookMarkInfo))
		aBookMarkInfo.m_RtfBookMark->ResetContent();
	else
		AfxMessageBox("RtfDoc::ClearBlock Import block: " + aBMImpBlkBegin + "not found");
}

CString RtfDoc::RtfToDosPath(const CString &aRtfPath)
{
	CString strFileName;
	for (int k=0; k < aRtfPath.GetLength(); k++)
	{
		if (aRtfPath[k] == '#')
			strFileName += '\\';
		else if (aRtfPath[k] == '@')
			strFileName += ':';
		else
			strFileName += aRtfPath[k];
	}
	return strFileName;
}

int RtfDoc::BlkImpFileOpen ( CString  &OtherImpBlkFileName)
{
//!    '*** Bestandsnaam van het blok vastleggen, bijv. BLDOSS.DOC ***
	int TempLengteBlokFileName = OtherImpBlkFileName.ReverseFind('\\');
    CString BlokFileName = OtherImpBlkFileName.Right(OtherImpBlkFileName.GetLength()- TempLengteBlokFileName - 1);
//!            MESSAGE ( ThisFuncName & |
//!                'BlokFileName ' & CLIP(BlokFileName))
 //!   '*** Directory van het hoofddocument vastleggen ***
    int TempLengteHoofdFileName = m_strFileName.ReverseFind('\\');
    CString HoofdFileDirectory = m_strFileName.Left(TempLengteHoofdFileName + 1);
	 HoofdFileDirectory += BlokFileName;

    if (OpenRtfDoc (HoofdFileDirectory)) //fail
	 {
        MESSAGE ( "RtfDoc::BlkImpFileOpen "
            "Opening HoofdFileDirectory " + HoofdFileDirectory + "failed" +
            "Will try to open with the original filename")
        if (OpenRtfDoc (OtherImpBlkFileName))
		  {
            MESSAGE ( "RtfDoc::BlkImpFileOpen "
            "Opening original OtherImpBlkFileName "+
            OtherImpBlkFileName + " also failed")
            return 1;
			}
        else
            return 0;
	 }
    else //ok
	 {
        OtherImpBlkFileName = HoofdFileDirectory;
        return 0;
    }
}

BOOL RtfDoc::IsSpace(const CString &str) const
{
	for (int k=0; k < str.GetLength(); k++)
		if (str[k] != ' ')
			return FALSE;

	return TRUE;
}

